from ..transformersx_base import *


@configclass
class ServiceArgs:
    model_name = 'bert-base-chinese'
    task_name = 'task'
    port = 9091
    host = '0.0.0.0'
    models_dir = '/app/models/finetuning'
    access_log = False
    debug = False



from transformers import BertModel

from transformersx.data import TaskDataProcessor
from transformersx.task import DefaultTransformerTask, TaskConfig
from transformersx.train import TaskContext
from transformersx.transformersx_base import to_json_string


class DefaultTransformerTaskForTest(DefaultTransformerTask):
    def __init__(self, config: TaskConfig):
        super().__init__(config)

    def _create_task_context(self, config: TaskConfig) -> TaskContext:
        data_processor = TaskDataProcessor()
        data_processor.get_labels = lambda: ['0']
        return TaskContext(
            task_name='sentiment',
            data_processor=data_processor,
            model_class=BertModel,
            compute_metrics=self.compute_metrics
        )

    def compute_metrics(self):
        pass


class Test_Transformersx_task:
    def test_task_config(self):
        print(to_json_string(TaskConfig()))

    def test_default_task(self):
        DefaultTransformerTaskForTest(TaskConfig())



from transformers import PreTrainedModel
import os
from pathlib import Path
import shutil
import re
from .trainer_base import *
from .trainer_optimizers import TaskTrainerOptimizers


@configclass
class TrainerCheckpointConfig():
    save_steps: int = field(500, "Save checkpoint every X updates steps.")
    checkpoint_base_dir: str = field('/app/models/finetuning',
                                     "The output directory where the model predictions and checkpoints will be written.")
    use_mtime: bool = field(False, 'Whether using mtime in checkpoint')
    checkpoint_prefix: str = field('checkpoint', 'the prefix of checkpoint dir')
    save_total_limit: int = field(None, "Limit the total amount of checkpoints."
                                        "Deletes the older checkpoints in the output_dir. Default is unlimited checkpoints")


class TaskTrainerCheckpoint():
    def __init__(self, trainer_env: TrainerEnv,
                 checkpoint_dir):
        self._env = trainer_env
        self.checkpoint_dir = checkpoint_dir

    def save(self, in_model, optimizers: TaskTrainerOptimizers):
        model = optimizers.get_model()
        if hasattr(model, "module"):
            assert model.module is in_model
        else:
            assert model is in_model

        self._save_check_point_model(in_model, self.checkpoint_dir)
        torch.save(optimizers.get_optimizer().state_dict(), os.path.join(self.checkpoint_dir, "optimizer.pt"))
        torch.save(optimizers.get_scheduler().state_dict(), os.path.join(self.checkpoint_dir, "scheduler.pt"))
        log.info("Saving optimizer and scheduler states to %s", self.checkpoint_dir)
        return self

    def load(self, optimizers: TaskTrainerOptimizers):
        if (os.path.isfile(os.path.join(self.checkpoint_dir, "optimizer.pt"))
                and os.path.isfile(os.path.join(self.checkpoint_dir, "scheduler.pt"))
        ):
            # Load in optimizer and scheduler states
            log.info("Loading optimizer and scheduler from :" + self.checkpoint_dir)
            optimizers.get_optimizer().load_state_dict(
                torch.load(os.path.join(self.checkpoint_dir, "optimizer.pt")))
            optimizers.get_scheduler().load_state_dict(
                torch.load(os.path.join(self.checkpoint_dir, "scheduler.pt")))
        return self

    def checkpoint_number(self):
        return int(self.checkpoint_dir.split("-")[-1].split("/")[0])

    def _save_check_point_model(self, model, output_dir: Optional[str] = None):
        os.makedirs(output_dir, exist_ok=True)

        if self._env.is_tpu_available():
            self._save_check_point_tpu(model, output_dir)

        elif self._env.is_world_master():
            self._save_check_point(model, output_dir)

    def _save_check_point_tpu(self, model, output_dir: Optional[str] = None):
        log.info("Saving model checkpoint to %s", output_dir)

        if self._env.is_master_ordinal():
            torch.save(self._env.args, os.path.join(output_dir, "training_args.bin"))

        # Save a trained model and configuration using `save_pretrained()`.
        # They can then be reloaded using `from_pretrained()`
        assert isinstance(model, PreTrainedModel), "Trainer.model appears to not be a PreTrainedModel"

        self._env.xm_rendezvous("saving_checkpoint")
        model.save_pretrained(output_dir)

    def _save_check_point(self, model, output_dir: Optional[str] = None):
        log.info("Saving model checkpoint to %s", output_dir)
        # Save a trained model and configuration using `save_pretrained()`.
        # They can then be reloaded using `from_pretrained()`
        assert isinstance(model, PreTrainedModel), "Trainer.model appears to not be a PreTrainedModel"

        model.save_pretrained(output_dir)

        # Good practice: save your training arguments together with the trained model
        torch.save(self._env.args, os.path.join(output_dir, "training_args.bin"))


class TaskTrainerCheckpointer():
    def __init__(self, task_name, trainer_env: TrainerEnv, config: TrainerCheckpointConfig):
        self._task_name = task_name
        self.config = config
        self._env = trainer_env
        self._model, self._optimizers = None, None

    def init_for_train(self, model: PreTrainedModel, trainer_optimizers: TaskTrainerOptimizers):
        self._model = model
        self._optimizers = trainer_optimizers

    def __make_checkpoint_dir(self, global_step):
        return join_path(
            self.config.checkpoint_base_dir, f"{self.config.checkpoint_prefix}-{global_step}"
        )

    def find_latest_checkpoint(self):
        checkpoints_sorted = self._sorted_checkpoints()
        if checkpoints_sorted:
            return TaskTrainerCheckpoint(self._env, join_path(self.config.checkpoint_base_dir, checkpoints_sorted[-1]))
        return None

    def save_check_point(self, global_step):
        assert self._model, 'init_for_train must be called before saving checkpoint'

        if self.config.save_steps > 0 and global_step % self.config.save_steps == 0:
            TaskTrainerCheckpoint(self._env,
                                  self.__make_checkpoint_dir(global_step)
                                  ).save(self._model, self._optimizers)
            self._rotate_checkpoints()

    def _sorted_checkpoints(self) -> List[str]:
        ordering_and_checkpoint_path = []

        glob_checkpoints = [str(x) for x in
                            Path(self.config.checkpoint_base_dir).glob(f"{self.config.checkpoint_prefix}-*")]

        for path in glob_checkpoints:
            if self.config.use_mtime:
                ordering_and_checkpoint_path.append((os.path.getmtime(path), path))
            else:
                regex_match = re.match(f".*{self.config.checkpoint_prefix}-([0-9]+)", path)
                if regex_match and regex_match.groups():
                    ordering_and_checkpoint_path.append((int(regex_match.groups()[0]), path))

        checkpoints_sorted = sorted(ordering_and_checkpoint_path)
        checkpoints_sorted = [checkpoint[1] for checkpoint in checkpoints_sorted]
        return checkpoints_sorted

    def _rotate_checkpoints(self) -> None:
        if self.config.save_total_limit is None or self.config.save_total_limit <= 0:
            return

        # Check if we should delete older checkpoint(s)
        checkpoints_sorted = self._sorted_checkpoints()
        if len(checkpoints_sorted) <= self.config.save_total_limit: return

        number_of_checkpoints_to_delete = max(0, len(checkpoints_sorted) - self.config.save_total_limit)
        checkpoints_to_be_deleted = checkpoints_sorted[:number_of_checkpoints_to_delete]
        for checkpoint in checkpoints_to_be_deleted:
            log.info("Deleting older checkpoint [{}] due to args.save_total_limit".format(checkpoint))
            shutil.rmtree(checkpoint)



from typing import Optional
import pandas as pd
from ..transformersx_base import join_path, log
from .data_models import TaskInputExample


class TaskDataProcessor:
    """Base class for data converters for sequence classification data sets."""

    def get_train_examples(self, limit_length: Optional[int] = None):
        """Gets a collection of `InputExample`s for the train set."""
        raise NotImplementedError()

    def get_eval_examples(self, limit_length: Optional[int] = None):
        """Gets a collection of `InputExample`s for the eval set."""
        raise NotImplementedError()

    def get_labels(self):
        """Gets the list of labels for this data set."""
        raise NotImplementedError()


class ValuesDataProcessor(TaskDataProcessor):
    def __init__(self, train_examples=None, eval_examples=None, labels=None):
        self.__train_examples, self.__eval_examples, self.__labels = train_examples, eval_examples, labels

    def get_train_examples(self, limit_length: Optional[int] = None):
        return self.__train_examples

    def get_eval_examples(self, limit_length: Optional[int] = None):
        return self.__eval_examples

    def get_labels(self):
        return self.__labels


class DefaultDataProcessor(TaskDataProcessor):
    def __init__(self, data_dir, labels=None, train_file='train.csv',
                 eval_file='dev.csv'):
        self._data_dir = data_dir
        self._labels = labels
        self._train_file = train_file
        self._eval_file = eval_file

    def _create_examples(self, file_name, type):
        return []

    def get_train_examples(self, limit_length: Optional[int] = None):
        return self._create_examples(self._train_file, 'train')

    def get_eval_examples(self, limit_length: Optional[int] = None):
        return self._create_examples(self._eval_file, 'dev')

    def get_labels(self):
        return self._labels


class CSVDataProcessor(DefaultDataProcessor):
    def __init__(self, data_dir, labels=None, label_col=0, text_a_col=1, text_b_col=-1, train_file='train.csv',
                 eval_file='dev.csv'):
        super().__init__(data_dir, labels, train_file, eval_file)
        self._label_col = label_col
        self._text_a_col = text_a_col
        self._text_b_col = text_b_col

    def _create_examples(self, file_name, type):
        pd_all = pd.read_csv(join_path(self._data_dir, file_name))

        log.info("Read data from {}, length={}".format(join_path(self._data_dir, file_name), len(pd_all)))
        examples = []
        for i, d in enumerate(pd_all.values):
            example = TaskInputExample(guid=type + '_' + str(i), text_a=d[self._text_a_col])
            if self._label_col > -1: example.label = str(d[self._label_col])
            if self._text_b_col > -1: example.text_b = d[self._text_b_col]

            examples.append(example)

        return examples



from ..transformersx_base import *


@configclass
class TaskDataConfig:
    data_dir: str = field('/app/dataset/datastore', 'input the data dir for processing')
    overwrite: bool = field(False, "Overwrite the cached training and evaluation sets")
    progress_bar: bool = field(default=True, metadata={"help": "Whether shows the progress_bar"})




from ai_harness.configuration import Arguments, configclass, field, ComplexArguments, export, merge_fields, \
    to_json_string
from ai_harness import configclasses
from ai_harness import harnessutils as aiutils
from ai_harness.fileutils import join_path, FileLineReader

log = aiutils.getLogger('transformersx')


@configclass()
class ArgumentsBase:
    pass



from tqdm import tqdm
from transformers import PreTrainedModel

from .trainer_base import *
from .trainer_dataloaders import TaskTrainerDataLoaders
from .trainer_metrics import TaskTrainerMetrics


@configclass
class TrainerEvaluatorConfig():
    pass


class TaskTrainerPredictor():
    def __init__(self, trainer_env: TrainerEnv, model: PreTrainedModel,
                 data_loaders: TaskTrainerDataLoaders):
        self._env = trainer_env
        self._model = model
        self._data_loaders = data_loaders

    def predict(self, test_dataset: Dataset) -> TaskPredictionOutput:
        test_dataloader = self._data_loaders.get_test_dataloader(test_dataset)
        return self.prediction_loop(test_dataloader, description="Prediction")

    def _prepare_model(self):
        # multi-gpu eval
        model = self._model
        if self._env.n_gpu > 1 and not isinstance(self._model, torch.nn.DataParallel):
            model = torch.nn.DataParallel(self._model)

        model.to(self._env.device)
        return model

    def _get_batch_size(self, dataloader: DataLoader):
        return dataloader._loader._loader.batch_size if self._env.is_tpu_available() else dataloader.batch_size

    def prediction_loop(
            self, dataloader: DataLoader, description: str, prediction_loss_only: Optional[bool] = None
    ) -> TaskPredictionOutput:
        prediction_loss_only = prediction_loss_only if prediction_loss_only is not None else prediction_loss_only
        model = self._prepare_model()
        batch_size = self._get_batch_size(dataloader)

        log.info("***** Running %s *****", description)
        log.info("  Num examples = %d", self._env.num_examples(dataloader))
        log.info("  Batch size = %d", batch_size)

        eval_losses: List[float] = []
        preds: np.ndarray = None
        label_ids: np.ndarray = None
        guids = []

        model.eval()

        for inputs in tqdm(dataloader, desc=description):

            has_labels = any(inputs.get(k) is not None for k in ["labels", "lm_labels", "masked_lm_labels"])

            has_guid = False
            for k, v in inputs.items():
                if k != 'guid':
                    inputs[k] = v.to(self._env.device)
                else:
                    has_guid = True
            if has_guid: guids.extend(inputs.pop('guid'))

            with torch.no_grad():
                outputs = model(**inputs)
                if has_labels:
                    step_eval_loss, logits = outputs[:2]
                    eval_losses += [step_eval_loss.mean().item()]
                else:
                    logits = outputs[0]

            if not prediction_loss_only:
                if preds is None:
                    preds = logits.detach().cpu().numpy()
                else:
                    preds = np.append(preds, logits.detach().cpu().numpy(), axis=0)
                if inputs.get("labels") is not None:
                    if label_ids is None:
                        label_ids = inputs["labels"].detach().cpu().numpy()
                    else:
                        label_ids = np.append(label_ids, inputs["labels"].detach().cpu().numpy(), axis=0)

        if self._env.is_tpu_available() and preds is not None and label_ids is not None:
            # tpu-comment: Get all predictions and labels from all worker shards of eval dataset
            preds, label_ids = self._env.get_tpu_eval(preds, label_ids)

        return TaskPredictionOutput(guids=guids, predictions=preds, label_ids=label_ids, eval_losses=eval_losses)


class TaskTrainerEvaluator():
    def __init__(self, trainer_env: TrainerEnv,
                 config: TrainerEvaluatorConfig,
                 predictor: TaskTrainerPredictor,
                 compute_metrics: TaskTrainerMetrics = None):
        self._env = trainer_env
        self.config = config
        self._predictor = predictor
        self._compute_metrics = compute_metrics

    def evaluate(self, dataloader: DataLoader, description: str, prediction_loss_only: Optional[bool] = None):
        p = self._predictor.prediction_loop(dataloader, description, prediction_loss_only)
        if self._compute_metrics: self._compute_metrics.eval_metrics(p)
        return p



from typing import List, Optional

from .data_converter import TaskDataConverter
from .data_processor import TaskDataProcessor

from .data_store import TaskDataStore, TaskDataset
from ..utils import torch_distributed_zero_first


class TaskDatasetFactory:
    def __init__(self, data_store: TaskDataStore,
                 processor: TaskDataProcessor = None,
                 dataConverter: TaskDataConverter = None):
        self._data_store = data_store
        self._processor = processor
        self._converter = dataConverter

    def _create_dataset(self, evaluate=False, limit_length: Optional[int] = None, local_rank=-1) -> TaskDataset:
        with torch_distributed_zero_first(local_rank):
            dataset = self._data_store.load_dataset(limit_length, evaluate)
            if not dataset:
                assert self._processor and self._converter, 'processor and converter must be set for no cache data'
                features = self.__generate_features(limit_length, evaluate)
                dataset = TaskDataset(features)
                if local_rank in [-1, 0]:
                    self._data_store.save_dataset(dataset, evaluate)
            return dataset

    def create_train_dataset(self, limit_length: Optional[int] = None, local_rank=-1) -> TaskDataset:
        return self._create_dataset(False, limit_length, local_rank)

    def create_eval_dataset(self, limit_length: Optional[int] = None, local_rank=-1) -> TaskDataset:
        return self._create_dataset(True, limit_length, local_rank)

    def __generate_features(self, limit_length: Optional[int] = None, evaluate=False):
        examples = (
            self._processor.get_eval_examples(limit_length) if evaluate
            else self._processor.get_train_examples(limit_length)
        )

        return self._converter.convert(examples)



from typing import Dict

from transformers.data.metrics import acc_and_f1

from transformersx.train import TaskPredictionOutput
from transformersx.train.trainer_metrics import TaskTrainerMetrics
from .news_data_processor import NewsDataProcessor, NewsDataConfig
import numpy as np

from ..task_base import log, TaskConfig, configclass, DefaultTransformerTask, TaskContext


@configclass
class NewsSegmentTaskArguments(TaskConfig):
    processor_config: NewsDataConfig = NewsDataConfig()


class NewsComputeMetrics(TaskTrainerMetrics):
    def eval_metrics(self, p: TaskPredictionOutput) -> Dict:
        preds = np.argmax(p.predictions, axis=1)
        result = acc_and_f1(preds, p.label_ids)
        log.info(p)
        log.info("preds.size=" + str(len(preds)) +
                 ", preds.sum=" + str(preds.sum()) +
                 ", label.sum=" + str(p.label_ids.sum()))

        correct = ((preds == 0) * (p.label_ids == 0)).sum()
        a1 = correct / (len(p.label_ids) - p.label_ids.sum())

        result['segment_acc'] = a1
        return result


class NewsSegmentTask(DefaultTransformerTask):
    args_class = NewsSegmentTaskArguments

    def __init__(self, config: NewsSegmentTaskArguments):
        super().__init__(config)

    def _create_task_context(self, config: NewsSegmentTaskArguments) -> TaskContext:
        return TaskContext(
            task_name='sentiment',
            data_processor=NewsDataProcessor(config.processor_config),
            compute_metrics=NewsComputeMetrics()
        )



from typing import Tuple

from transformers.file_utils import cached_property
from torch.utils.data.dataloader import DataLoader
from torch.utils.data.dataset import Dataset
from torch.utils.data.distributed import DistributedSampler
from torch.utils.data.sampler import RandomSampler
from .trainer_utils import *
from ..transformersx_base import log, configclass, field, join_path, export

try:
    import torch_xla.core.xla_model as xm

    _has_tpu = True
except ImportError:
    _has_tpu = False

if _has_tpu:
    import torch_xla.core.xla_model as xm
    import torch_xla.debug.metrics as met
    import torch_xla.distributed.parallel_loader as pl

try:
    from apex import amp

    _has_apex = True
except ImportError:
    _has_apex = False


@configclass
class TrainerEvnConfig():
    local_rank: int = field(-1, "For distributed training: local_rank")
    tpu_num_cores: Optional[int] = field(
        None, "TPU: Number of TPU cores (automatically passed by launcher script)")
    no_cuda: bool = field(False, "Do not use CUDA even when it is available")
    seed: int = field(42, "random seed for initialization")

    fp16: bool = field(
        False,
        "Whether to use 16-bit (mixed) precision (through NVIDIA apex) instead of 32-bit")
    fp16_opt_level: str = field(
        "O1",
        "For fp16: Apex AMP optimization level selected in ['O0', 'O1', 'O2', and 'O3']."
        "See details at https://nvidia.github.io/apex/amp.html")

    per_gpu_train_batch_size: int = field(16, "Batch size per GPU/CPU for training.")
    per_gpu_eval_batch_size: int = field(16, "Batch size per GPU/CPU for evaluation.")


class TrainerEnv:
    def __init__(self, args=None):
        self.args = args
        self.config: TrainerEvnConfig = self.get_config(TrainerEvnConfig)

    def get_config(self, config_class):
        if not self.args: return config_class()
        return export(self.args, config_class)

    def is_tpu_available(self):
        return _has_tpu

    def is_apex_available(self):
        return _has_apex

    def get_tpu_sampler(self, dataset: Dataset):
        if xm.xrt_world_size() <= 1:
            return RandomSampler(dataset)
        return DistributedSampler(dataset, num_replicas=xm.xrt_world_size(), rank=xm.get_ordinal())

    def xrt_world_size(self):
        return xm.xrt_world_size()

    def is_local_master(self) -> bool:
        if self.is_tpu_available():
            return xm.is_master_ordinal(local=True)
        else:
            return self.config.local_rank in [-1, 0]

    def get_tpu_dataloader(self, data_loader):
        return pl.ParallelLoader(data_loader, [self.device]).per_device_loader(self.device)

    def tpu_metrics(self):
        xm.master_print(met.metrics_report())

    def apex_model_optimizer(self, model, optimizer):
        assert self.is_apex_available(), "Please install apex from https://www.github.com/nvidia/apex to use fp16 training."
        return amp.initialize(model, optimizer, opt_level=self.config.fp16_opt_level)

    def get_tpu_eval(self, preds, label_ids):
        preds = xm.mesh_reduce("eval_preds", preds, np.concatenate)
        label_ids = xm.mesh_reduce("eval_out_label_ids", label_ids, np.concatenate)
        return preds, label_ids

    def is_master_ordinal(self):
        return xm.is_master_ordinal()

    def xm_rendezvous(self, tag):
        xm.rendezvous(tag)

    def fp16_backward(self, optimizer, loss):
        with amp.scale_loss(loss, optimizer) as scaled_loss:
            scaled_loss.backward()

    def fp16_clip(self, optimizer, max_grad_norm):
        torch.nn.utils.clip_grad_norm_(amp.master_params(optimizer), max_grad_norm)

    def tpu_step(self, optimizer):
        xm.optimizer_step(optimizer)

    def is_world_master(self) -> bool:
        """
        This will be True only in one process, even in distributed mode,
        even when training on multiple machines.
        """
        if self.is_tpu_available():
            return xm.is_master_ordinal(local=False)
        else:
            return self.config.local_rank == -1 or torch.distributed.get_rank() == 0

    def num_examples(self, dataloader: Union[DataLoader, "pl.PerDeviceLoader"]) -> int:
        """
        Helper to get num of examples from a DataLoader, by accessing its Dataset.
        """
        if self.is_tpu_available():
            assert isinstance(dataloader, pl.PerDeviceLoader)
            return len(dataloader._loader._loader.dataset)
        else:
            return len(dataloader.dataset)

    def batch_train_size(self):
        return self.config.per_gpu_train_batch_size * max(1, self.n_gpu)

    def batch_eval_size(self):
        return self.config.per_gpu_eval_batch_size * max(1, self.n_gpu)

    @cached_property
    def _setup_devices(self) -> Tuple["torch.device", int]:
        log.info("PyTorch: setting up devices")
        if self.config.no_cuda:
            device = torch.device("cpu")
            n_gpu = 0
        elif self.is_tpu_available():
            device = xm.xla_device()
            n_gpu = 0
        elif self.config.local_rank == -1:
            # if n_gpu is > 1 we'll use nn.DataParallel.
            # If you only want to use a specific subset of GPUs use `CUDA_VISIBLE_DEVICES=0`
            device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
            n_gpu = torch.cuda.device_count()
        else:
            # Here, we'll use torch.distributed.
            # Initializes the distributed backend which will take care of sychronizing nodes/GPUs
            torch.distributed.init_process_group(backend="nccl")
            device = torch.device("cuda", self.config.local_rank)
            n_gpu = 1
        return device, n_gpu

    @property
    def device(self) -> "torch.device":
        return self._setup_devices[0]

    @property
    def n_gpu(self):
        return self._setup_devices[1]



from collections import OrderedDict

from transformers import is_tf_available
from transformers.modeling_electra import ElectraConfig

from transformers.tokenization_electra import ElectraTokenizer, ElectraTokenizerFast
from .modelingx_electra import ElectraForSequenceClassificationX

from ..model_base import TaskModels, ModelType, ModelTaskType, model_func

default_model = model_func(ModelType.electra, ElectraConfig, ElectraTokenizer, 'electra')
fast_model = model_func(ModelType.electra, ElectraConfig, ElectraTokenizerFast, 'electra')


class Electra_Task_Models_Base(TaskModels):
    MODELS = {
        "cn": [default_model("hfl/chinese-electra-small-discriminator"),
               default_model("hfl/chinese-electra-small-generator"),
               default_model("hfl/chinese-electra-base-discriminator"),
               default_model("hfl/chinese-electra-base-generator"),
               default_model("hfl/chinese-electra-large-discriminator"),
               default_model("hfl/chinese-electra-large-generator")
               ]}


class Electra_Task_Models(Electra_Task_Models_Base):
    from transformers.modeling_electra import ElectraModel, ElectraForMaskedLM, ElectraForPreTraining, \
        ElectraDiscriminatorPredictions, ElectraForTokenClassification, ElectraGeneratorPredictions, \
        ElectraForSequenceClassification
    MODEL_CLASSES = OrderedDict([
        (ModelTaskType.base, ElectraModel),
        (ModelTaskType.pretrain, ElectraForPreTraining),
        (ModelTaskType.lm_head, ElectraForMaskedLM),
        (ModelTaskType.seq_cls, ElectraForSequenceClassificationX),
        (ModelTaskType.token_cls, ElectraForTokenClassification)
    ])


if is_tf_available():
    class TFElectra_Task_Models(Electra_Task_Models_Base):
        from transformers.modeling_tf_electra import TFElectraModel, TFElectraForMaskedLM, TFElectraForPreTraining, \
            TFElectraDiscriminatorPredictions, TFElectraForTokenClassification, TFElectraGeneratorPredictions
        MODEL_CLASSES = OrderedDict([
            (ModelTaskType.base, TFElectraModel),
            (ModelTaskType.pretrain, TFElectraForPreTraining),
            (ModelTaskType.lm_head, TFElectraForMaskedLM),
            (ModelTaskType.token_cls, TFElectraForTokenClassification)
        ])



import numpy as np
import torch
import random
from contextlib import contextmanager


def set_seed(seed: int):
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    # ^^ safe to call this function even if cuda is not available


@contextmanager
def torch_distributed_zero_first(local_rank: int):
    """
    Decorator to make all processes in distributed training wait for the first one (locally) to do something.
    """
    if local_rank not in [-1, 0]:
        torch.distributed.barrier()
    yield
    if local_rank == 0:
        torch.distributed.barrier()



import random

from tqdm import tqdm

from transformersx.data.data_processor import DefaultDataProcessor, ValuesDataProcessor
from transformersx.utils import cut_sentences
from ..task_base import *


@configclass
class NewsDataConfig:
    raw_data_dir: str = field('/app/dataset/news', 'input the data dir for processing')
    save_mid: bool = field(True, 'whether cache the middle data for check')
    context_min_len: int = field(64, 'context min length')
    sentence_min_len: int = field(10, 'sentence min length')
    check_min_anyway: int = field(True, 'whether check the sentence min length anyway')
    positive_mode: int = field(0, 'positive mode')
    negative_mode: int = field(0, 'negative mode')
    bar_size: int = field(1000, 'the progress bar size')


class SentenceSplitter:
    def __init__(self, sentence_min_len=10, check_min_anyway=False):
        self._min_len = sentence_min_len
        self._check_min_anyway = check_min_anyway

    def split(self, segment: str):
        t = segment.replace('\n', '')

        sentences = cut_sentences(t)
        length = len(sentences)
        sents = []
        i = 1
        l1 = len(sentences[i - 1])
        while i < length:
            l2 = len(sentences[i])
            if l1 > self._min_len or (l1 != l2 and not self._check_min_anyway):
                sents.append(sentences[i - 1])
                if i == length - 1: sents.append(sentences[i])
                i = i + 1
                l1 = l2
                continue
            start = i - 1
            while True:
                if i == length or sentences[i - 1][-1] != sentences[i][-1]: break
                i = i + 1

            sents.append(''.join(sentences[start:i]))
            i = i + 1
            l1 = l2
        return sents


class SentencesSegment():
    def __init__(self, segment: str = None, sentence_min_len=10, exclude_empty_line=True, check_min_anyway=False):
        self._exclude_empty_line = exclude_empty_line
        self._sentences = self._make_sentences(segment, sentence_min_len, check_min_anyway)
        self._size = len(self._sentences)

    def _make_sentences(self, segment: str, sentence_min_len, check_min_anyway=False):
        if not segment: return []
        return SentenceSplitter(sentence_min_len, check_min_anyway).split(segment)

    def add_new_sentence(self, sentence: str):
        sentence = sentence.strip()
        if self._exclude_empty_line and not sentence: return
        self._sentences.append(sentence)
        self._size = self._size + 1
        return self

    def add_new_sentences_not_split(self, sentences):
        if not sentences: return

        self.add_new_sentences(sentences.split('\n'))
        return self

    def add_new_sentences(self, sentences):
        for line in sentences: self.add_new_sentence(line)
        return self

    def sentences(self, from_index, to_index):
        if to_index > self._size: to_index = self._size
        if from_index < 0: from_index = 0
        if from_index >= to_index: return None

        return ''.join(self._sentences[from_index:to_index])

    def first_sentences(self, len):
        if len < 1: return None
        len = self._size if len > self._size else len
        return self.sentences(0, len)

    def first_sentence(self):
        return self._sentences[0]

    def sentence(self, s_i):
        if s_i >= self._size: return None
        return self._sentences[s_i]

    def last_sentence(self):
        return self._sentences[-1]

    def size(self):
        return self._size

    def char_index(self, sentence_index):
        index = 0
        for i in range(sentence_index): index = index + len(self._sentences[i])
        return index

    def char_indexes(self, sentences_indexes):
        if not sentences_indexes:  return []
        sentences_indexes.sort()
        from_index = 0
        return_indexes = []
        cur_index = 0
        for i in range(len(sentences_indexes)):
            for j in range(from_index, sentences_indexes[i] + 1):
                cur_index = cur_index + len(self._sentences[j])
            return_indexes.append(cur_index)
            from_index = sentences_indexes[i] + 1

        return return_indexes

    def content(self, separate_indexes):
        indexes = [*separate_indexes, len(self._sentences) - 1]
        indexes.sort()
        from_index = 0
        return_content = [''] * len(indexes)
        for i in range(len(indexes)):
            for j in range(from_index, indexes[i] + 1):
                return_content[i] = return_content[i] + self._sentences[j]
            from_index = indexes[i] + 1
        return return_content

    def all_sentences(self):
        return self._sentences

    def all_content(self):
        return ''.join(self._sentences)


class NewsExampleSegment(SentencesSegment):
    def __init__(self, segment: str, context_min_len=64, sentence_min_len=10, exclude_empty_line=True,
                 check_min_anyway=False):
        super().__init__(segment, sentence_min_len, exclude_empty_line, check_min_anyway)
        self._context_min_len = context_min_len
        self.contexts = []
        self._make_contexts()

    @staticmethod
    def from_article(article: str, sentence_line=False):
        if sentence_line:
            segment = NewsExampleSegment('').add_new_sentences_not_split(article)
            segment._make_contexts()
            return segment
        return NewsExampleSegment(article)

    def _make_contexts(self):
        total = self.size()
        if total < 1:  return []

        length = len(self.first_sentence())
        start = 1
        while start < total and length < self._context_min_len:
            length = length + len(self.sentence(start))
            start = start + 1
        self.contexts.append((0, start))

        last_start = 0
        length = length - len(self.sentence(last_start))
        for end in range(start + 1, total):
            length = length + len(self.sentence(end))
            if length < self._context_min_len: continue

            last_start = last_start + 1
            self.contexts.append((last_start, end + 1))
            length = length - len(self.sentence(last_start))

    def context(self, c_start, c_end):
        return self.sentences(c_start, c_end)

    def last_context(self):
        return self.sentences(self.contexts[-1][0], self.contexts[-1][1])

    def example(self, c_start, c_end):
        return self.context(c_start, c_end), self.sentence(c_end)

    @staticmethod
    def from_file(file_path, context_min_len=64, sentence_min_len=10,
                  exclude_empty_line=True,
                  check_min_anyway=False, line_sentence=True):
        if line_sentence:
            segment = NewsExampleSegment(None, context_min_len, sentence_min_len, exclude_empty_line, check_min_anyway)
            FileLineReader(bar_step_size=-1, exclude_empty_line=True).pipe(
                lambda input, result: segment.add_new_sentence(input)
            ).read(file_path)
            segment._make_contexts()
            return segment

        content = []
        FileLineReader(bar_step_size=-1, exclude_empty_line=True).pipe(
            lambda input, result: content.append(input)
        ).read(file_path)
        return NewsExampleSegment("".join(content), context_min_len, sentence_min_len,
                                  exclude_empty_line,
                                  check_min_anyway)


class NewsExampleGenerator():
    def __init__(self, config: NewsDataConfig, type='train'):
        self._type = type
        self._config = config
        self.examples = []
        self._last_segment: NewsExampleSegment = None
        self._example_id = 0

    def add_line(self, new_segment_str: str):
        new_segment = NewsExampleSegment(new_segment_str,
                                         context_min_len=self._config.context_min_len,
                                         sentence_min_len=self._config.sentence_min_len)
        if not new_segment.size(): return self

        if self._last_segment is not None:
            # In order to balance the number of classification,
            if eval('self._create_positive_examples_' + str(self._config.positive_mode) + '()'):
                eval('self._create_negative_examples_' + str(self._config.negative_mode) + '(new_segment)')

        self._last_segment = new_segment

        return self

    def _guid(self):
        self._example_id = self._example_id + 1
        return "%s-%s" % (self._type, self._example_id)

    def _create_positive_examples_0(self):
        if len(self._last_segment.contexts) < 2: return False
        e_index = random.choice(self._last_segment.contexts[:-1])
        guid = self._guid()
        text_a, text_b = self._last_segment.example(*e_index)
        self.examples.append(TaskInputExample(guid=guid, text_a=text_a, text_b=text_b, label='0'))
        return True

    def _create_positive_examples_1(self):
        for e_index in self._last_segment.contexts[:-1]:
            guid = self._guid()
            text_a, text_b = self._last_segment.example(*e_index)
            self.examples.append(TaskInputExample(guid=guid, text_a=text_a, text_b=text_b, label='0'))

    def _create_negative_examples_0(self, new_segment: NewsExampleSegment):
        guid = self._guid()
        text_a = self._last_segment.last_context()
        text_b = new_segment.sentence(0)
        self.examples.append(TaskInputExample(guid=guid, text_a=text_a, text_b=text_b, label='1'))

    def _create_negative_examples_1(self, new_segment: NewsExampleSegment):
        guid = self._guid()
        for context in self._last_segment.contexts:
            text_a = self._last_segment.context(*context)
            for text_b in new_segment.all_sentences():
                self.examples.append(TaskInputExample(guid=guid, text_a=text_a, text_b=text_b, label='1'))


class FileNewsExampleProcessor:
    def __init__(self, file, config: NewsDataConfig, type='train'):
        self._file = file
        self._config = config
        self._type = type
        self._example_generator = NewsExampleGenerator(config, type)

    def _make_examples(self):
        def handle_line(line, previous):
            self._example_generator.add_line(line)

        return handle_line

    def _save_middle_data(self):
        log.info("Save the middle data: ")
        with open(join_path(self._config.raw_data_dir, self._type + "_middle_examples.txt"), 'w') as f:
            for example in tqdm(self._example_generator.examples, disable=(self._config.bar_size <= 0)):
                f.write(example.text_a + '\n')
                f.write(example.text_b + '\n')
                f.write(example.label + '\n\n')

    def get_examples(self):
        FileLineReader(self._config.bar_size).pipe(self._make_examples()).read(self._file)
        if self._config.save_mid: self._save_middle_data()
        return self._example_generator.examples


class NewsDataProcessor(DefaultDataProcessor):
    def __init__(self, config: NewsDataConfig):
        super().__init__(config.raw_data_dir, labels=['0', '1'], train_file='train.txt', eval_file='dev.txt')
        self._config = parse_tasks_args(NewsDataConfig) if not config else config

    def _create_examples(self, file_name, type):
        return FileNewsExampleProcessor(join_path(self._config.raw_data_dir, file_name), self._config,
                                        type).get_examples()


class PredictDataProcessor(ValuesDataProcessor):
    def __init__(self, segment: NewsExampleSegment):
        super().__init__(eval_examples=self._create_example_from_article(segment), labels=['0', '1'])

    @staticmethod
    def _create_example_from_article(segment: NewsExampleSegment):
        examples = []
        for i, e_index in enumerate(segment.contexts[:-1]):
            text_a, text_b = segment.example(*e_index)
            examples.append(TaskInputExample(guid=str(i), text_a=text_a, text_b=text_b))
        return examples



import numpy as np
from sklearn.metrics import f1_score
from transformers.data.metrics import simple_accuracy

from .. import *


class ExampleTaskBase:
    args_class = TaskArguments

    def __init__(self, task_name, taskArgs: TaskArguments = None, task_model_class=None):
        self._task_name = task_name
        if taskArgs is not None:
            self.task_args = taskArgs
        else:
            self.task_args = parse_tasks_args()
        self._task_model_class = task_model_class

    def __build_task(self):
        return TransformerTask(self._task_name, self.task_args, self._data_processor(),
                               model_class=self._task_model_class,
                               compute_metric=self._compute_metrics)

    def train(self):
        self.task_args.training_args.train()
        self.__build_task().train()

    def eval(self):
        self.task_args.training_args.do_eval = True
        self.__build_task().eval()

    def _acc_and_f1(self, preds, labels):
        acc = simple_accuracy(preds, labels)
        f1 = f1_score(y_true=labels, y_pred=preds, average="weighted")
        return {
            "acc": acc,
            "f1": f1,
            "acc_and_f1": (acc + f1) / 2,
        }

    def _compute_metrics(self, p: EvalPrediction):
        preds = np.argmax(p.predictions, axis=1)
        result = self._acc_and_f1(preds, p.label_ids)
        log.info(p)
        return result

    def _data_processor(self):
        raise NotImplementedError



from ..transformersx_base import configclass, field
from .models import ALL_TASK_MODEL_PATHS
from .model_base import ModelTaskType, ModelMode


@configclass()
class TaskModelConfig:
    model_pretrained_dir: str = field("/app/models/pretrained",
                                      "This folder is for the pretrained models downloaded from Internet.")
    model_name: str = field("bert-base-chinese", "the name of model: " + str(ALL_TASK_MODEL_PATHS))
    model_task_type: str = field("seq_cls",
                                 "the task type of model:{}, model_task_type decides the Task".format(
                                     str(ModelTaskType.names())))
    model_mode: str = field("classification", "the model of model: " + str(ModelMode.names()))
    framework: str = field("pt", "the name of framework: pt--Pytorch,tf--Tensorflow")
    language: str = field("cn", "the language of model: cn, en")
    num_labels: int = field(2, "the number of label")
    max_len: int = field(512, "the max length of model input")
    unit_test: bool = field(False, "For unit test of the code, without the load of the model")
    freeze_parameters: str = field("bert", "The parameters list for freeze")
    freeze_main: bool = field(False, 'whether freeze main parameter')



from tqdm import tqdm, trange
from transformers import DataCollator

from .trainer_base import *
from .trainer_config import TrainerConfig
from .trainer_metrics import TaskTrainerMetrics
from .trainer_scheduler import TaskTrainedScheduler
from .trainer_logger import TaskTrainerLogger
from .trainer_optimizers import TaskTrainerOptimizers
from .trainer_checkpoint import TaskTrainerCheckpointer
from .trainer_dataloaders import TaskTrainerDataLoaders
from .trainer_evaluator import TaskTrainerEvaluator, TaskTrainerPredictor
from ..utils import set_seed
from ..model import TaskModel
from ..data import TaskDatasetFactory, TaskDataConverter, TaskDataProcessor
from ..transformersx_base import field, configclass
import os


@configclass
class TaskContext:
    task_name: str = field('default')
    data_collator: DataCollator = field(TaskDefaultDataCollatorx())
    compute_metrics: TaskTrainerMetrics = None
    model_class: type = None
    data_processor: TaskDataProcessor = None


class TaskTrainerBuildContext:
    def task_context(self) -> TaskContext: raise NotImplementedError()

    def task_model(self) -> TaskModel: raise NotImplementedError()

    def task_dataset_factory(self) -> TaskDatasetFactory: raise NotImplementedError()

    def task_data_converter(self) -> TaskDataConverter: raise NotImplementedError()

    def activate_context(self, model_path=None, for_train=True): raise NotImplementedError()


class TaskTrainerContextBase():
    def __init__(self, train_env: TrainerEnv, config: TrainerConfig, build_context: TaskTrainerBuildContext,
                 model_path=None, for_train=False):
        self.env = train_env
        self.config = config
        self.task_model, self.model = self.__activate_context_and_build_model(build_context, model_path, for_train)
        self.task_context, self.dataset_factory = build_context.task_context(), build_context.task_dataset_factory()
        self.task_name = self.task_context.task_name
        self.data_loaders = TaskTrainerDataLoaders(self.env, self.config.dl_config, self.dataset_factory,
                                                   self.task_context.data_collator)
        self.predictor = TaskTrainerPredictor(self.env, self.model, self.data_loaders)

    def __activate_context_and_build_model(self, build_context: TaskTrainerBuildContext, model_path, for_train):
        build_context.activate_context(model_path, for_train)
        task_model = build_context.task_model()
        self.__init_model_env(task_model.model)

        return task_model, task_model.model

    def __init_model_env(self, model):
        model.to(self.env.device)
        set_seed(self.env.config.seed)

        if self.env.is_tpu_available():
            # Set an xla_device flag on the model's config.
            # We'll find a more elegant and not need to do this in the future.
            model.config.xla_device = True


class TaskTrainerContext_Eval(TaskTrainerContextBase):
    def __init__(self, train_env: TrainerEnv,
                 config: TrainerConfig,
                 build_context: TaskTrainerBuildContext, model_path=None):
        super().__init__(train_env, config, build_context, model_path, False)
        self.evaluator = TaskTrainerEvaluator(self.env, self.config.eval_config, self.predictor,
                                              self.task_context.compute_metrics)


class TaskTrainerContext_Train(TaskTrainerContextBase):
    def __init__(self, train_env: TrainerEnv,
                 config: TrainerConfig,
                 build_context: TaskTrainerBuildContext, model_path=None):
        self.trainer_checkpointer = TaskTrainerCheckpointer(build_context.task_context().task_name, train_env,
                                                            config.chk_config)
        latest_checkpoint = self.trainer_checkpointer.find_latest_checkpoint()
        model_path = latest_checkpoint.checkpoint_dir if latest_checkpoint else model_path
        super().__init__(train_env, config, build_context, model_path, True)

        self.evaluator = TaskTrainerEvaluator(self.env, self.config.eval_config, self.predictor,
                                              self.task_context.compute_metrics)

        self.optimizers, self.train_counter = self.__create_optimizers_counter(self.model, latest_checkpoint)
        self.trainer_logger = TaskTrainerLogger(self.env, self.model)

    def __create_optimizers_counter(self, task_model, latest_checkpoint):
        self.config.sch_config.global_step = 0 if not latest_checkpoint else latest_checkpoint.checkpoint_number()
        self.config.sch_config.train_data_len = len(self.data_loaders.get_train_dataloader())
        self.config.sch_config.train_batch_size = self.data_loaders.train_batch_size
        train_scheduler = TaskTrainedScheduler(self.env, self.config.sch_config)
        optimizers = TaskTrainerOptimizers(self.env, task_model, train_scheduler.num_train_epochs)

        self.trainer_checkpointer.init_for_train(task_model, optimizers)
        optimizers.set_for_start_train()
        return optimizers, train_scheduler


class TaskTrainStep():
    def __init__(self, trainer_env: TrainerEnv, taskContext: TaskTrainerContext_Train, total_steps):
        self._env = trainer_env
        self._context = taskContext
        self._total_steps = total_steps

    def execute(self, step, inputs: Dict[str, torch.Tensor]) -> (float, bool):
        loss = self._context.optimizers.step_backward(inputs)
        if not self._context.train_counter.is_step(step, self._total_steps): return loss, False
        self._context.optimizers.clip_and_step()
        return loss, True


class TaskTrainEpoch():
    def __init__(self, epoch, taskContext: TaskTrainerContext_Train):
        self._env = taskContext.env
        self._context = taskContext
        self.config = taskContext.config
        self.epoch_iterator = tqdm(taskContext.data_loaders.get_train_dataloader(), desc="Iteration",
                                   disable=not self._env.is_local_master())
        self.epoch = epoch
        self.epoch_len = len(self.epoch_iterator)
        self._train_counter = taskContext.train_counter
        self._trainer_step = self.__create_step()

    def __create_step(self):
        return TaskTrainStep(self._env, self._context, self.epoch_len)

    def execute(self):
        for step, inputs in enumerate(self.epoch_iterator):
            if not self._train_counter.check_step_for_break(self.epoch_iterator): break
            if self._train_counter.check_step_for_skip(): continue

            loss, do_step = self._trainer_step.execute(step, inputs)
            self._process_step(step, do_step, loss)

        self.epoch_iterator.close()
        self._context.trainer_logger.log_train_epoch()

    def _process_step(self, step, do_step, loss):
        self._train_counter.on_train_step(step, self.epoch_len, loss)

        if do_step and self._env.is_local_master():
            self._process_step_loss(self.epoch, self._train_counter.tr_loss, self._train_counter.logging_loss)
            self._train_counter.after_process_step()

    def _process_step_loss(self, epoch, tr_loss, logging_loss):
        if self._context.trainer_logger.is_need_log_step(self._train_counter.global_step):
            logs = {}
            logs["loss"] = (tr_loss - logging_loss) / self.config.log_config.logging_steps
            logs["learning_rate"] = self._context.optimizers.scheduler.get_last_lr()[0]

            self._context.trainer_logger.log_train_step(epoch, self._train_counter.global_step, logs)

            if self.config.training_config.evaluate_during_training:
                self._context.evaluator.evaluate(self._context.data_loaders.get_eval_dataloader(),
                                                 description="Training setup evaluate, step={}".format(
                                                     self._train_counter.global_step))

        self._context.trainer_checkpointer.save_check_point(self._train_counter.global_step)


class TaskTrainLoop():
    def __init__(self, taskContext: TaskTrainerContext_Train):
        self._env = taskContext.env
        self.training_config = taskContext.config.training_config
        self.task_context = taskContext
        self._train_counter = taskContext.train_counter

    def execute(self):
        train_iterator = trange(
            self._train_counter.epochs_trained, int(self._train_counter.num_train_epochs), desc="Epoch",
            disable=not self._env.is_local_master()
        )

        for epoch in train_iterator:
            if not self._train_counter.check_step_for_break(train_iterator):
                break
            TaskTrainEpoch(epoch, self.task_context).execute()

        train_iterator.close()
        self.task_context.trainer_logger.log_train_end()
        self.__save_trained_model()
        return self._train_counter.task_train_output()

    def __save_trained_model(self):
        os.makedirs(self.training_config.output_dir, exist_ok=True)
        self.task_context.model.save_pretrained(self.training_config.output_dir)
        if self._env.is_world_master():
            self.task_context.task_model.tokenizer.save_pretrained(self.training_config.output_dir)


class TaskTrainer():
    def __init__(self, trainer_env: TrainerEnv, build_context: TaskTrainerBuildContext):
        self._env = trainer_env
        self.config = TrainerConfig.from_env(trainer_env)
        self._build_context = build_context

    def train(self, model_path=None):
        context = TaskTrainerContext_Train(self._env, self.config, self._build_context, model_path)
        context.trainer_logger.log_pre_train(context.data_loaders.get_train_dataloader(), context.train_counter)
        return TaskTrainLoop(context).execute()

    def evaluate(self, model_path=None):
        context = TaskTrainerContext_Eval(self._env, self.config, self._build_context, model_path)
        return context.evaluator.evaluate(context.data_loaders.get_eval_dataloader(), description="Evaluate")

    def predictor(self, model_path=None) -> TaskTrainerPredictor:
        predict_context = TaskTrainerContextBase(self._env, self.config, self._build_context, model_path, False)
        return predict_context.predictor



from transformers import PreTrainedModel, get_linear_schedule_with_warmup, AdamW
from .trainer_base import *


@configclass()
class TrainerOptimizersConfig:
    learning_rate: float = field(5e-5, "The initial learning rate for Adam.")
    weight_decay: float = field(0.0, "Weight decay if we apply some.")
    adam_epsilon: float = field(1e-8, "Epsilon for Adam optimizer.")
    max_grad_norm: float = field(1.0, "Max gradient norm.")
    warmup_steps: int = field(0, "Linear warmup over warmup_steps.")
    gradient_accumulation_steps: int = field(
        1, "Number of updates steps to accumulate before performing a backward/update pass.")


class TaskTrainerOptimizers():
    def __init__(self, trainer_env: TrainerEnv,
                 model: PreTrainedModel,
                 num_train_steps: int,
                 optimizer: torch.optim.Optimizer = None,
                 scheduler: torch.optim.lr_scheduler.LambdaLR = None
                 ):
        self._env = trainer_env
        self.config: TrainerOptimizersConfig = trainer_env.get_config(TrainerOptimizersConfig)
        self.origin_model = model
        self.model = model
        self._optimizer = optimizer
        self._scheduler = scheduler
        self._num_train_steps = num_train_steps

    @property
    def optimizer(self) -> torch.optim.Optimizer:
        if not self._optimizer:
            self._optimizer = self._create_optimizer()
        return self._optimizer

    @property
    def scheduler(self) -> torch.optim.lr_scheduler.LambdaLR:
        if not self._scheduler:
            self._scheduler = self._create_scheduler(self.optimizer, self._num_train_steps)

        return self._scheduler

    def set_for_start_train(self):
        if self._env.config.fp16:
            self.model, self._optimizer = self._env.apex_model_optimizer(self.model, self.optimizer)

            # multi-gpu training (should be after apex fp16 initialization)
        if self._env.n_gpu > 1:
            self.model = torch.nn.DataParallel(self.model)

        # Distributed training (should be after apex fp16 initialization)
        if self._env.config.local_rank != -1:
            self.model = torch.nn.parallel.DistributedDataParallel(
                self.model,
                device_ids=[self._env.config.local_rank],
                output_device=self._env.config.local_rank,
                find_unused_parameters=True
            )
        self.model.zero_grad()

    def _create_optimizer(self):
        no_decay = ["bias", "LayerNorm.weight"]
        optimizer_grouped_parameters = [
            {
                "params": [p for n, p in self.model.named_parameters() if not any(nd in n for nd in no_decay)],
                "weight_decay": self.config.weight_decay,
            },
            {
                "params": [p for n, p in self.model.named_parameters() if any(nd in n for nd in no_decay)],
                "weight_decay": 0.0,
            },
        ]
        return AdamW(optimizer_grouped_parameters, lr=self.config.learning_rate, eps=self.config.adam_epsilon)

    def _create_scheduler(self, optimizer, num_training_steps):
        return get_linear_schedule_with_warmup(
            optimizer, num_warmup_steps=self.config.warmup_steps, num_training_steps=num_training_steps
        )

    def step_backward(self, inputs: Dict[str, torch.Tensor]) -> float:
        self.model.train()
        for k, v in inputs.items():
            inputs[k] = v.to(self._env.device)

        outputs = self.model.forward(**inputs)
        loss = outputs[0]  # model outputs are always tuple in transformers (see doc)

        if self._env.n_gpu > 1:
            loss = loss.mean()  # mean() to average on multi-gpu parallel training
        if self.config.gradient_accumulation_steps > 1:
            loss = loss / self.config.gradient_accumulation_steps

        if self._env.config.fp16:
            self._env.fp16_backward(loss, self.optimizer)
        else:
            loss.backward()

        return loss.item()

    def clip_and_step(self):
        if self._env.config.fp16:
            self._env.fp16_clip(self.optimizer, self.config.max_grad_norm)
        else:
            torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.config.max_grad_norm)

        if self._env.is_tpu_available():
            self._env.tpu_step(self.optimizer)
        else:
            self.optimizer.step()

        self.scheduler.step()
        self.model.zero_grad()



from ..task_base import *


class NewsClassificationTask(DefaultTransformerTask):
    def __init__(self, config: TaskConfig):
        super().__init__(config)

    def _create_task_context(self, config: TaskConfig) -> TaskContext:
        return TaskContext(
            task_name='news_cls',
            data_processor=CSVDataProcessor(
                data_dir=config.data_config.data_dir,
                labels=["体育", "娱乐", "家居", "房产", "教育", "时尚", "时政", "游戏", "科技", "财经"]
            ))



from .trainer_base import *


@configclass()
class TrainerSchedulerConfig:
    max_steps: int = field(
        -1, "If > 0: set total number of training steps to perform. Override num_train_epochs.")
    num_train_epochs: float = field(3.0, "Total number of training epochs to perform.")

    # not field
    train_data_len = 0
    gradient_accumulation_steps = 1
    train_batch_size = 16
    global_step = 0


@configclass()
class TaskTrainedScheduler():
    epoch = 0
    global_step = 0
    epochs_trained = 0
    steps_trained_in_current_epoch = 0

    tr_loss = 0.0
    logging_loss = 0.0

    total_steps = 0
    num_train_epochs = 0
    total_train_batch_size = 0

    def __init__(self, trainer_env: TrainerEnv, config: TrainerSchedulerConfig):
        self._env = trainer_env
        self.config = config
        self.global_step = config.global_step

        self._cal_epochs()
        self._cal_steps()

    def _cal_epochs(self):
        if self.config.max_steps > 0:
            self.t_total = self.config.max_steps
            self.num_train_epochs = (
                    self.config.max_steps // (self.config.train_data_len // self.config.gradient_accumulation_steps) + 1
            )
        else:
            self.t_total = int(
                self.config.train_data_len // self.config.gradient_accumulation_steps * self.config.num_train_epochs)
            self.num_train_epochs = self.config.num_train_epochs

        if self._env.is_tpu_available():
            self.total_train_batch_size = self.config.train_batch_size * self._env.xrt_world_size()
        else:
            self.total_train_batch_size = (
                    self.config.train_batch_size
                    * self.config.gradient_accumulation_steps
                    * (torch.distributed.get_world_size() if self._env.config.local_rank != -1 else 1)
            )

    def _cal_steps(self):
        if self.global_step == 0:
            return
        try:
            self.epochs_trained = self.global_step // (
                    self.config.train_data_len // self.config.gradient_accumulation_steps)
            self.steps_trained_in_current_epoch = self.global_step % (
                    self.config.train_data_len // self.config.gradient_accumulation_steps
            )

            log.info("  Continuing training from checkpoint, will skip to saved global_step")
            log.info("  Continuing training from epoch %d", self.epochs_trained)
            log.info("  Continuing training from global step %d", self.global_step)
            log.info("  Will skip the first %d steps in the first epoch", self.steps_trained_in_current_epoch)
        except ValueError:
            self.global_step = 0

    def on_train_step(self, step, epoch_len, loss):
        self.global_step += 1
        self.epoch = self.epoch + (step + 1) / epoch_len
        self.tr_loss += loss

    def after_process_step(self):
        self.logging_loss = self.tr_loss

    def check_step_for_break(self, iterator):
        if self.total_steps <= 0 or self.global_step < self.total_steps:
            return True
        iterator.close()
        return False

    def check_step_for_skip(self):
        if self.steps_trained_in_current_epoch <= 0:
            return False
        self.steps_trained_in_current_epoch -= 1
        return True

    def task_train_output(self):
        TaskTrainOutput(self.global_step, self.tr_loss / self.global_step)

    def is_step(self, step, total_steps):
        return (step + 1) % self.config.gradient_accumulation_steps == 0 or (
            # last step in epoch but step is always smaller than gradient_accumulation_steps
                total_steps <= self.config.gradient_accumulation_steps
                and (step + 1) == total_steps
        )



from typing import Optional, Union

from transformers.tokenization_utils import BatchEncoding

from .data_models import TaskInputExample, TaskInputFeatures
from transformers import PreTrainedTokenizer
from ..transformersx_base import log
from tqdm.auto import tqdm, trange


class TaskDataConverter:
    def convert(self, examples):
        raise NotImplementedError()


class DefaultTaskDataConverter(TaskDataConverter):
    def __init__(self, tokenizer: PreTrainedTokenizer, label_list, max_length: Optional[int] = None,
                 progress_bar=False):
        self._tokenizer = tokenizer
        self._progress_bar = progress_bar
        self._max_length = max_length if max_length else tokenizer.max_len
        if self._max_length > 1024: self._max_length = 512
        self._label_list = label_list
        self._label_map = {label: i for i, label in enumerate(self._label_list)} if label_list else None

    def _label_from_example(self, example: TaskInputExample) -> Union[int, float]:
        if self._label_map:
            return self._label_map[example.label]
        else:
            return float(example.label)

    def _batch_encode_plus(self, examples):
        log.info("1. Tokenizer encoding examples .... total: " + str(len(examples)))
        total = len(examples)
        epoch_iterator = tqdm(range(0, total, 100), desc="Iteration", disable=not self._progress_bar)

        batch_outputs = {}
        for step in epoch_iterator:
            batch_encoding = self._tokenizer.batch_encode_plus(
                [(example.text_a, example.text_b) for example in examples[step:step + 100]],
                max_length=self._max_length,
                pad_to_max_length=True,
            )

            for key, value in batch_encoding.items():
                if key not in batch_outputs:
                    batch_outputs[key] = []
                batch_outputs[key].extend(value)

        return BatchEncoding(batch_outputs)

    def _convert_features(self, examples, labels, batch_encoding):
        log.info("2. Converting Examples to Features .... total: " + str(len(examples)))
        epoch_iterator = tqdm(examples, desc="Iteration", disable=not self._progress_bar)
        features = []

        for k in batch_encoding.keys():
            log.info("key={},size={}".format(k, str(len(batch_encoding[k]))))

        for i, example in enumerate(epoch_iterator):
            inputs = {k: batch_encoding[k][i] for k in batch_encoding}
            feature = TaskInputFeatures(**inputs, label=labels[i] if labels else None, guid=i if not labels else None)
            features.append(feature)
        return features

    def convert(self, examples):
        has_label = (examples[0].label is not None)
        labels = [self._label_from_example(example) for example in examples] if has_label else None
        batch_encoding = self._batch_encode_plus(examples)

        features = self._convert_features(examples, labels, batch_encoding)

        for i, example in enumerate(examples[:5]):
            log.info("*** Example ***")
            log.info("guid: %s" % (example.guid))
            log.info("features: %s" % features[i])

        return features



import json
from typing import Dict, NamedTuple, Optional, List, Union

import dataclasses
import numpy as np
import torch
from dataclasses import dataclass
from transformers.data.data_collator import InputDataClass


@dataclass
class InputExample:
    """
    A single training/test example for simple sequence classification.

    Args:
        guid: Unique id for the example.
        text_a: string. The untokenized text of the first sequence. For single
            sequence tasks, only this sequence must be specified.
        text_b: (Optional) string. The untokenized text of the second sequence.
            Only must be specified for sequence pair tasks.
        label: (Optional) string. The label of the example. This should be
            specified for train and dev examples, but not for test examples.
    """

    guid: str
    text_a: str
    text_b: Optional[str] = None
    label: Optional[str] = None

    def to_json_string(self):
        """Serializes this instance to a JSON string."""
        return json.dumps(dataclasses.asdict(self), indent=2) + "\n"


@dataclass(frozen=True)
class InputFeatures:
    """
    A single set of features of data.
    Property names are the same names as the corresponding inputs to a model.

    Args:
        input_ids: Indices of input sequence tokens in the vocabulary.
        attention_mask: Mask to avoid performing attention on padding token indices.
            Mask values selected in ``[0, 1]``:
            Usually  ``1`` for tokens that are NOT MASKED, ``0`` for MASKED (padded) tokens.
        token_type_ids: (Optional) Segment token indices to indicate first and second
            portions of the inputs. Only some models use them.
        label: (Optional) Label corresponding to the input. Int for classification problems,
            float for regression problems.
    """

    input_ids: List[int]
    attention_mask: Optional[List[int]] = None
    token_type_ids: Optional[List[int]] = None
    label: Optional[Union[int, float]] = None
    guid: int = None

    def to_json_string(self):
        """Serializes this instance to a JSON string."""
        return json.dumps(dataclasses.asdict(self)) + "\n"


class EvalPrediction(NamedTuple):
    """
    Evaluation output (always contains labels), to be used
    to compute metrics.
    """

    predictions: np.ndarray
    label_ids: np.ndarray


class PredictionOutput(NamedTuple):
    guids: np.ndarray
    predictions: np.ndarray
    label_ids: Optional[np.ndarray]
    metrics: Optional[Dict[str, float]]


class TrainOutput(NamedTuple):
    global_step: int
    training_loss: float


class DefaultDataCollatorx:
    """
    Very simple data collator that:
    - simply collates batches of dict-like objects
    - Performs special handling for potential keys named:
        - `label`: handles a single value (int or float) per object
        - `label_ids`: handles a list of values per object
    - does not do any additional preprocessing

    i.e., Property names of the input object will be used as corresponding inputs to the model.
    See glue and ner for example of how it's useful.
    """

    def __call__(self, features: List[InputDataClass]) -> Dict[str, torch.Tensor]:
        # In this method we'll make the assumption that all `features` in the batch
        # have the same attributes.
        # So we will look at the first element as a proxy for what attributes exist
        # on the whole batch.
        first = features[0]

        # Special handling for labels.
        # Ensure that tensor is created with the correct type
        # (it should be automatically the case, but let's make sure of it.)
        if hasattr(first, "label") and first.label is not None:
            if type(first.label) is int:
                labels = torch.tensor([f.label for f in features], dtype=torch.long)
            else:
                labels = torch.tensor([f.label for f in features], dtype=torch.float)
            batch = {"labels": labels}
        elif hasattr(first, "label_ids") and first.label_ids is not None:
            if type(first.label_ids[0]) is int:
                labels = torch.tensor([f.label_ids for f in features], dtype=torch.long)
            else:
                labels = torch.tensor([f.label_ids for f in features], dtype=torch.float)
            batch = {"labels": labels}
        elif hasattr(first, 'guid') and first.guid is not None:
            guids = torch.tensor([f.guid for f in features], dtype=torch.long)
            batch = {"guid": guids}
        else:
            batch = {}

        # Handling of all other possible attributes.
        # Again, we will use the first element to figure out which key/values are not None for this model.
        for k, v in vars(first).items():
            if k not in ("guid", "label", "label_ids") and v is not None and not isinstance(v, str):
                batch[k] = torch.tensor([getattr(f, k) for f in features], dtype=torch.long)
        return batch


PREFIX_CHECKPOINT_DIR = "checkpoint"



from transformers import DataCollator

from .trainer_base import *
from ..data import TaskDatasetFactory


@configclass
class TrainerDataloadersConfig():
    pass


class TaskTrainerDataLoaders():
    def __init__(self, trainer_env: TrainerEnv,
                 config: TrainerDataloadersConfig,
                 dataset_factory: TaskDatasetFactory,
                 data_collator: DataCollator):
        self._env = trainer_env
        self.config = config
        self._data_collator = data_collator
        self._dataset_factory = dataset_factory
        self._local_rank = trainer_env.config.local_rank
        self.train_batch_size = trainer_env.batch_train_size()
        self.eval_batch_size = trainer_env.batch_eval_size()

    def _create_dataloader(self, dataset, sampler, batch_size):
        data_loader = DataLoader(
            dataset,
            sampler=sampler,
            batch_size=batch_size,
            shuffle=False if sampler is not None else True,
            collate_fn=self._data_collator,
        )
        if self._env.is_tpu_available():
            data_loader = self._env.get_tpu_dataloader(data_loader)
        return data_loader

    def get_train_dataloader(self) -> DataLoader:
        train_dataset = self._dataset_factory.create_train_dataset(local_rank=self._local_rank) \
            if self._dataset_factory else None
        assert train_dataset, '"Trainer: training requires a train_dataset."'
        if self._env.is_tpu_available():
            train_sampler = self._env.get_tpu_sampler(train_dataset)
        else:
            train_sampler = (
                RandomSampler(train_dataset)
                if self._local_rank == -1
                else DistributedSampler(train_dataset)
            )

        return self._create_dataloader(train_dataset, train_sampler, self.train_batch_size)

    def get_eval_dataloader(self, eval_dataset: Optional[Dataset] = None) -> DataLoader:
        eval_dataset = self._dataset_factory.create_eval_dataset(
            local_rank=self._local_rank) if not eval_dataset and self._dataset_factory else eval_dataset
        assert eval_dataset, "Trainer: evaluation requires an eval_dataset."

        sampler = self._env.get_tpu_sampler(eval_dataset) if self._env.is_tpu_available() else None

        return self._create_dataloader(eval_dataset, sampler, self.eval_batch_size)

    def get_test_dataloader(self, test_dataset: Dataset) -> DataLoader:
        # We use the same batch_size as for eval.
        sampler = None
        if self._env.is_tpu_available():
            sampler = self._env.get_tpu_sampler(test_dataset) if self._env.is_tpu_available() else None

        return self._create_dataloader(test_dataset, sampler, self.eval_batch_size)



import time
from typing import Optional, List

import torch

from torch.utils.data import Dataset
import os

from .data_models import TaskInputFeatures
from .data_config import TaskDataConfig
from ..transformersx_base import log, join_path


class TaskDataset(Dataset):
    def __init__(self, features: List[TaskInputFeatures]):
        self.features = features

    def __len__(self):
        return len(self.features)

    def __getitem__(self, i) -> TaskInputFeatures:
        return self.features[i]


class TaskDataStore():
    dataStoreId = None
    dataArgs = None

    def load_dataset(self, limit_length: Optional[int] = None, evaluate=False) -> TaskDataset:
        raise NotImplementedError()

    def save_dataset(self, dataset: TaskDataset, evaluate=False):
        raise NotImplementedError()


class LocalDataStore(TaskDataStore):
    def __init__(self, dataStoreId, config: TaskDataConfig):
        self.dataStoreId = dataStoreId
        self.config = config

    def load_dataset(self, limit_length: Optional[int] = None, evaluate=False) -> TaskDataset:
        saved_file = self._create_save_file(evaluate)
        if not os.path.exists(saved_file):
            return None
        log.info("Loading dataset from file {}".format(saved_file))
        start = time.time()
        features = torch.load(saved_file)
        log.info(
            f"Loaded dataset from file %s [took %.3f s]", saved_file, time.time() - start
        )
        return TaskDataset(features)

    def _create_save_file(self, evaluate):
        saved_file = "{}_{}".format(self.dataStoreId, ("eval" if evaluate else "train"))
        return join_path(self.config.data_dir, saved_file)

    def save_dataset(self, dataset: TaskDataset, evaluate=False):
        if not dataset:
            return
        saved_file = self._create_save_file(evaluate)
        if not os.path.exists(saved_file) or self.config.overwrite:
            log.info("Saving dataset into  file {}".format(saved_file))
            start = time.time()
            torch.save(dataset.features, saved_file)

            log.info(
                f"Saved dataset into cached file %s [took %.3f s]", saved_file, time.time() - start
            )



from ..train import TrainerEnv
from ..model import TaskModelFactory
from ..train import TaskContext, TaskTrainerBuildContext, TaskTrainer
from ..data import TaskDatasetFactory, LocalDataStore, TaskDataConverter, DefaultTaskDataConverter
from .task_config import TaskConfig


class DefaultTaskTrainerBuildContext(TaskTrainerBuildContext):
    def __init__(self, task_config: TaskConfig, task_context: TaskContext):
        self.config = task_config
        self._task_context = task_context
        self._task_model, self._task_data_converter, self._task_dataset_factory = None, None, None
        self.config.model_config.num_labels = len(self._task_context.data_processor.get_labels())

    def task_context(self) -> TaskContext:
        return self._task_context

    def task_model(self) -> TaskModelFactory:
        assert self._task_model, 'call activate_context to activate context firstly'
        return self._task_model

    def __get_data_store_id(self):
        return "cached_{}_{}_{}".format(
            self._task_model.tokenizer.__class__.__name__,
            str(self.config.model_config.max_len),
            self._task_context.task_name,
        )

    def task_dataset_factory(self) -> TaskDatasetFactory:
        if self._task_dataset_factory: return self._task_dataset_factory
        assert self._task_model, 'call activate_context to activate context firstly'
        data_store_id = self.__get_data_store_id()
        task_data_store = LocalDataStore(data_store_id, self.config.data_config)

        self._task_dataset_factory = TaskDatasetFactory(task_data_store,
                                                        self._task_context.data_processor,
                                                        self.task_data_converter())
        return self._task_dataset_factory

    def task_data_converter(self) -> TaskDataConverter:
        if self._task_data_converter: return self._task_data_converter
        assert self._task_model, 'call activate_context to activate context firstly'
        self._task_data_converter = DefaultTaskDataConverter(
            self._task_model.tokenizer,
            self._task_context.data_processor.get_labels(),
            self.config.model_config.max_len,
            self.config.data_config.progress_bar
        )
        return self._task_data_converter

    def activate_context(self, model_path=None, for_train=True):
        if self._task_model: return self._task_model
        self._task_model = TaskModelFactory(self._task_context.task_name,
                                            self.config.model_config,
                                            self._task_context.model_class).get_task_model(model_path, for_train)
        return self._task_model


class TransformerTask:
    args_class = TaskConfig

    def train(self): raise NotImplementedError()

    def eval(self): raise NotImplementedError()

    def predict(self, examples): raise NotImplementedError()


class DefaultTransformerTask(TransformerTask):
    args_class = TaskConfig

    def __init__(self, config: TaskConfig, task_context: TaskContext = None):
        self.config = config
        self._task_context = self._create_task_context(config) if not task_context else task_context
        self._build_context = DefaultTaskTrainerBuildContext(self.config, self._task_context)
        self._task_trainer, self._task_predictor = None, None

    def _create_task_context(self, config: TaskConfig) -> TaskContext:
        raise NotImplementedError('this method must be implemented for None task_context argument in Construction')

    def _get_task_trainer(self) -> TaskTrainer:
        if not self._task_trainer: self._task_trainer = TaskTrainer(TrainerEnv(self.config.training_config),
                                                                    self._build_context)
        return self._task_trainer

    def train(self):
        self._get_task_trainer().train()

    def eval(self):
        return self._get_task_trainer().evaluate()

    def predict(self, examples):
        if not self._task_predictor:
            self._task_predictor = self._get_task_trainer().predictor()
        return self._task_predictor.predict(self._build_context.task_data_converter().convert(examples))



from typing import Optional
from unittest.mock import patch

from torch.utils.data import Dataset
from transformers.modeling_bert import BertConfig, BertModel, BertForSequenceClassification
from transformers.tokenization_bert import BertTokenizer

from transformersx.data import TaskDatasetFactory, TaskDataConverter, DefaultTaskDataConverter, TaskDataset, \
    TaskInputExample
from transformersx.model import TaskModel
from transformersx.train import TaskTrainerBuildContext, TaskContext, TaskTrainer, TrainerEnv


def create_test_examples(type, count):
    return [TaskInputExample(guid=str(i), text_a=type + '_text_a_' + str(i), text_b=type + '_text_b_' + str(i),
                             label=str(i % 2)) for i in range(count)]


class TaskDatasetFactoryForTest(TaskDatasetFactory):
    def __init__(self, converter: TaskDataConverter):
        self._converter = converter

    def create_eval_dataset(self, limit_length: Optional[int] = None, local_rank=-1) -> Dataset:
        return TaskDataset(self._converter.convert(create_test_examples('eval', 2)))

    def create_train_dataset(self, limit_length: Optional[int] = None, local_rank=-1) -> Dataset:
        return TaskDataset(self._converter.convert(create_test_examples('train', 5)))


class TaskTrainerBuildContextForTest(TaskTrainerBuildContext):
    def __init__(self):
        self.model = self._create_task_model()
        self.converter = DefaultTaskDataConverter(self.model.tokenizer, label_list=['0', '1'], max_length=128)

    def _create_task_model(self):
        bert_config = BertConfig(vocab_size=21128, hidden_size=312, num_hidden_layers=1, num_attention_heads=4,
                                 intermediate_size=1536, max_position_embeddings=128)
        bert_tokenizer = BertTokenizer('../dataset/vocab.txt')
        bert_model = BertForSequenceClassification(bert_config)
        return TaskModel(bert_config, bert_tokenizer, bert_model)

    def task_context(self) -> TaskContext:
        return TaskContext()

    def task_dataset_factory(self) -> TaskDatasetFactory:
        return TaskDatasetFactoryForTest(self.converter)

    def task_data_converter(self) -> TaskDataConverter:
        return DefaultTaskDataConverter(self.model.tokenizer, label_list=['0', '1'])

    def task_model(self) -> TaskModel:
        return self.model

    def activate_context(self, model_path=None, for_train=True):
        pass


class Test_Task_Trainer:

    def test_predictor(self):
        assert TaskTrainer(TrainerEnv(), TaskTrainerBuildContextForTest()).predictor()

    def test_predictor_with_model_path(self):
        assert TaskTrainer(TrainerEnv(), TaskTrainerBuildContextForTest()).predictor('/')

    def test_evaluate(self):
        TaskTrainer(TrainerEnv(), TaskTrainerBuildContextForTest()).evaluate()

    def test_evaluate_with_model_path(self):
        TaskTrainer(TrainerEnv(), TaskTrainerBuildContextForTest()).evaluate('/')

    def test_train(self):
        trainer = TaskTrainer(TrainerEnv(), TaskTrainerBuildContextForTest())
        trainer.config.training_config.output_dir = 'build'
        TaskTrainer(TrainerEnv(), TaskTrainerBuildContextForTest()).train()

    def test_train_with_model_path(self):
        trainer = TaskTrainer(TrainerEnv(), TaskTrainerBuildContextForTest())
        trainer.config.training_config.output_dir = 'build'
        TaskTrainer(TrainerEnv(), TaskTrainerBuildContextForTest()).train('/')



from ..transformersx_base import *
from . import ALL_TASK_MODEL_PATHS, ModelTaskType, ModelMode


@configclass
class ModelArguments:
    """
    Arguments pertaining to which model/config/tokenizer we are going to fine-tune from.
    """
    model_finetuned_dir: str = field("/app/models/finetuning",
                                     "The 'model_base_dir' generally is for the finetuned models."
                                     "Generally, the model is loaded from 'model_finetuned_dir' firstly."
                                     " If the model cannot be found, it will be loaded from the model_pretrained_dir.")

    model_pretrained_dir: str = field("/app/models/pretrained",
                                      "This folder is for the pretrained models downloaded from Internet.")

    not_use_pretrained: bool = field(False,
                                     "Whether use the model in model_pretrained_dir. Generally, for training, it should be False")

    model_name: str = field("bert-base-chinese", "the name of model: " + str(ALL_TASK_MODEL_PATHS))

    model_task_type: str = field("seq_cls",
                                 "the task type of model:{}, model_task_type decides the Task".format(
                                     str(ModelTaskType.names())))

    model_mode: str = field("classification", "the model of model: " + str(ModelMode.names()))

    framework: str = field("pt", "the name of framework: pt--Pytorch,tf--Tensorflow")

    language: str = field("cn", "the language of model: cn, en")

    freeze_parameters: str = field("bert", "The parameters list for freeze")
    freeze_main: bool = field(False, 'whether freeze main parameter')

    num_labels: int = field(2, "the number of label")

    unit_test: bool = field(False, "For unit test of the code, without the load of the model")



import json

from .trainer_base import TrainerEnv, TaskEvalPrediction, TaskPredictionOutput
from typing import Dict, NamedTuple, Optional, List, Union
import numpy as np

from transformers.data.metrics import acc_and_f1


class TaskTrainerMetrics:
    def eval_metrics(self, p: TaskPredictionOutput): raise NotImplementedError()


class DefaultTaskTrainerMetrics(TaskTrainerMetrics):
    def __init__(self, trainer_env: TrainerEnv, for_cls=True):
        self._env = trainer_env
        self._for_cls = for_cls

    def eval_metrics(self, p: TaskPredictionOutput):
        if p.predictions is not None and p.label_ids is not None:
            metrics = self._compute_metrics(TaskEvalPrediction(predictions=p.predictions, label_ids=p.label_ids))
        else:
            metrics = {}

        if len(p.eval_losses) > 0:
            metrics["eval_loss"] = np.mean(p.eval_losses)

        # Prefix all keys with eval_
        for key in list(metrics.keys()):
            if not key.startswith("eval_"):
                metrics[f"eval_{key}"] = metrics.pop(key)
        p.metrics = metrics

        output = json.dumps({**metrics})
        print(output)

        if self._env.args.tpu_metrics_debug:
            self._env.tpu_metrics()

    def _compute_metrics(self, p: TaskEvalPrediction) -> Dict:
        if self._for_cls:
            preds = np.argmax(p.predictions, axis=1)
        else:
            preds = np.squeeze(p.predictions)
        return acc_and_f1(preds, p.label_ids)



from unittest.mock import patch
from transformersx.model.model_factory import TaskModelFactory, TaskModelConfig


class Test_Task_Model_Factory:
    @patch('transformersx.model.model_base.TaskModel')
    def test_load_model(self, mock_class):
        instance = mock_class.return_value
        instance.load.return_value = (None, None, None)
        model_factory = TaskModelFactory('test_task', TaskModelConfig())
        assert not model_factory.get_task_model().config

    @patch('transformersx.model.model_base.TaskModel')
    def test_load_model_cached(self, mock_class):
        instance = mock_class.return_value
        instance.load.return_value = ('1', None, None)
        model = TaskModelFactory('test_task', TaskModelConfig()).get_task_model()
        assert model.config == '1'
        instance.load.return_value = (None, None, None)
        assert model.config == '1'
        model = TaskModelFactory('test_task', TaskModelConfig()).get_task_model('test')
        assert not model.config



from unittest.mock import patch
from .mock_class_b import Mock_Class_A, Mock_Class_B


def test_mock_class_a():
    assert Mock_Class_B().do()
    with patch('tests.transformersx.mock_class_b.Mock_Class_A') as Mock_Class:
        instance = Mock_Class.return_value
        instance.do.return_value = False
        assert Mock_Class_B().do() == False


@patch('tests.transformersx.mock_class_b.Mock_Class_A')
def test_mock_annotation(Mock_Class):
    assert Mock_Class_B().do()
    instance = Mock_Class.return_value
    instance.do.return_value = False
    assert Mock_Class_B().do() == False


class Test_Mock:
    @patch('tests.transformersx.mock_class_b.Mock_Class_A')
    def test_mock_annotation(self, Mock_Class):
        assert Mock_Class_B().do()
        instance = Mock_Class.return_value
        instance.do.return_value = False
        assert Mock_Class_B().do() == False



from transformersx import TransformerTask, ComplexArguments, log, configclass, field


@configclass()
class WorkspaceConfig:
    workspace_dir: str = field('/app/workspace', 'Transformers workspace directory')


class TransformersWorkspace:
    TASKS = dict()

    def __init__(self, config: WorkspaceConfig):
        self.config = config

    def register_task(self, task_name, task_class):
        assert not self.TASKS.get(task_name), "task name {} already exists, please use other name.".format(task_name)

        self.TASKS.setdefault(task_name, task_class)
        assert issubclass(task_class, (TransformerTask)), "task_class must be the subclass of TransformerTask"

        return self

    def register_tasks(self, tasks):
        if not tasks:
            return
        for task in tasks:
            self.register_task(task[0], task[1])
        return self

    def _build_arguments(self):
        arguments = {}
        for task_name, task_class in self.TASKS.items():
            arg_obj = task_class.args_class()
            arguments[task_name] = arg_obj
        return arguments

    def start_task(self, args=None, test=False):
        argument_objs = self._build_arguments()

        task_name, arguments = ComplexArguments(argument_objs).parse(args)
        if test: arguments.model_args.unit_test = test
        log.info("Start Example Task:{}, with arguments:{}".format(task_name, str(arguments)))
        task_action = arguments.action
        task_instance = self.TASKS[task_name](arguments)

        eval('task_instance.' + task_action + '()')



import json

from ..model.model_config import TaskModelConfig
from ..data.data_config import TaskDataConfig
from ..train.trainer_config import TaskTrainingConfig
from ..transformersx_base import *


def to_json_string(obj):
    """
    Serializes this instance to a JSON string.
    """
    return json.dumps(configclasses.asdict(obj), indent=2)


@configclass
class TaskConfig:
    action: str = field("train", "the task action: train,eval,predict")
    base_dir: str = field('/app/workspace', 'task base directory')
    model_config: TaskModelConfig = TaskModelConfig()
    data_config: TaskDataConfig = TaskDataConfig()
    training_config: TaskTrainingConfig = TaskTrainingConfig()

    def to_json_string(self):
        """
        Serializes this instance to a JSON string.
        """
        return json.dumps(configclasses.asdict(self), indent=2)


def parse_tasks_args(argsObjOrClass=None):
    if argsObjOrClass is None:
        task_args = TaskConfig()
    elif argsObjOrClass is not None and type(argsObjOrClass) != type:
        task_args = argsObjOrClass
    else:
        task_args = argsObjOrClass()

    return Arguments(task_args).parse()



from collections import OrderedDict

from transformers import is_tf_available

from .bert.models import Bert_Task_Models
from .albert.models import Albert_Task_Models
from .electra.models import Electra_Task_Models
from .roberta.models import Roberta_Task_Models
from .xlnet.models import XLNet_Task_Models
from .model_base import ModelType

if is_tf_available():
    from .bert.models import TFBert_Task_Models
    from .albert.models import TFAlbert_Task_Models
    from .electra.models import TFElectra_Task_Models
    from .roberta.models import TFRoberta_Task_Models
    from .xlnet.models import TFXLNet_Task_Models

    TF_ALL_TASK_MODELS = OrderedDict([
        (ModelType.albert, TFAlbert_Task_Models()),
        (ModelType.bert, TFBert_Task_Models()),
        (ModelType.roberta, TFRoberta_Task_Models()),
        (ModelType.electra, TFElectra_Task_Models()),
        (ModelType.xlnet, TFXLNet_Task_Models())
    ])

ALL_TASK_MODELS = OrderedDict([
    (ModelType.albert, Albert_Task_Models()),
    (ModelType.bert, Bert_Task_Models()),
    (ModelType.roberta, Roberta_Task_Models()),
    (ModelType.electra, Electra_Task_Models()),
    (ModelType.xlnet, XLNet_Task_Models())
])

ALL_TASK_MODEL_PATHS = [task_model.all_paths() for task_model in ALL_TASK_MODELS.values()]

FRAEMEWORKS = {}

FRAEMEWORKS.setdefault("pt", ALL_TASK_MODELS)
if is_tf_available():
    FRAEMEWORKS.setdefault("tf", TF_ALL_TASK_MODELS)


def query_base_models(types, framework='pt', language='cn'):
    f_models = FRAEMEWORKS.get(framework)

    if not f_models:
        return None
    return dict([(type, f_models.get(type).all_base_models(language)) for type in types if f_models.get(type)])


def task_model(model_path, model_task_type, model_type=None, language='cn', framework='pt'):
    f_models = FRAEMEWORKS.get(framework)
    if not f_models:
        raise ValueError(
            "Cannot find the model, path={}, language={}, model_task_type={}, framework={}, Framework is not support: {}".format(
                model_path, language, model_task_type, framework, framework))
    models = f_models.get(model_type)
    if models:
        return models.task_model(model_path, model_task_type, language, framework)
    return _task_model_no_type(model_path, model_task_type, language, framework)


def _task_model_no_type(model_path, model_task_type, language='cn', framework='pt'):
    for models in FRAEMEWORKS.get(framework).values():
        model = models.task_model(model_path, model_task_type, language, ignore_not_exists=True)
        if model:
            return model
    raise ValueError("Cannot find models with model path {}".format(model_path))



from .data_config import TaskDataConfig
from .data_processor import TaskDataProcessor, CSVDataProcessor
from .data_store import TaskDataset, LocalDataStore
from .dataset_factory import TaskDatasetFactory
from .data_converter import TaskDataConverter, DefaultTaskDataConverter
from .data_models import TaskInputExample, TaskInputFeatures



import json
import logging
from typing import Dict, Any

import torch
from transformers import is_torch_available

from ..transformersx_base import *
from .trainer_base import TrainerEvnConfig, TrainerEnv
from .trainer_checkpoint import TrainerCheckpointConfig
from .trainer_scheduler import TrainerSchedulerConfig
from .trainer_dataloaders import TrainerDataloadersConfig
from .trainer_evaluator import TrainerEvaluatorConfig
from .trainer_logger import TrainerLoggerConfig
from .trainer_optimizers import TrainerOptimizersConfig

logger = logging.getLogger(__name__)


class TrainerTrainingConfig:
    evaluate_during_training: bool = field(True, "Run evaluation during training at each logging step.")
    output_dir: str = field('/app/models/finetuning',
                            "The output directory where the model predictions and checkpoints will be written.")
    overwrite_output_dir: bool = field(
        False, "Overwrite the content of the output directory."
               "Use this to continue training if output_dir points to a checkpoint directory.")


@configclass()
class TrainerConfig:
    env_config: TrainerEvnConfig = field(TrainerEvnConfig())
    dl_config: TrainerDataloadersConfig = field(TrainerDataloadersConfig())
    eval_config: TrainerEvaluatorConfig = field(TrainerEvaluatorConfig())
    opt_config: TrainerOptimizersConfig = field(TrainerOptimizersConfig())
    chk_config: TrainerCheckpointConfig = field(TrainerCheckpointConfig())
    log_config: TrainerLoggerConfig = field(TrainerLoggerConfig())
    sch_config: TrainerSchedulerConfig = field(TrainerSchedulerConfig())
    training_config: TrainerTrainingConfig = field(TrainerTrainingConfig())

    def to_json_string(self):
        """
        Serializes this instance to a JSON string.
        """
        return json.dumps(configclasses.asdict(self), indent=2)

    def to_sanitized_dict(self) -> Dict[str, Any]:
        """
        Sanitized serialization to use with TensorBoard’s hparams
        """
        d = configclasses.asdict(self)
        valid_types = [bool, int, float, str]
        if is_torch_available():  valid_types.append(torch.Tensor)
        return {k: v if type(v) in valid_types else str(v) for k, v in d.items()}

    @staticmethod
    def from_env(env: TrainerEnv):
        return TrainerConfig(
            env_config=env.config,
            dl_config=env.get_config(TrainerDataloadersConfig),
            eval_config=env.get_config(TrainerEvaluatorConfig),
            opt_config=env.get_config(TrainerOptimizersConfig),
            chk_config=env.get_config(TrainerCheckpointConfig),
            log_config=env.get_config(TrainerLoggerConfig),
            sch_config=env.get_config(TrainerSchedulerConfig)
        )


@configclass()
class ConfigClass:
    pass


TaskTrainingConfig = merge_fields(ConfigClass, TrainerEvnConfig,
                                  TrainerOptimizersConfig,
                                  TrainerLoggerConfig, TrainerSchedulerConfig,
                                  TrainerDataloadersConfig,
                                  TrainerEvaluatorConfig, TrainerCheckpointConfig)



from transformers.tokenization_bert import BertTokenizer

from transformersx.data import *


def create_test_examples(type, count):
    return [TaskInputExample(guid=str(i), text_a=type + '_text_a_' + str(i), text_b=type + '_text_b_' + str(i),
                             label=str(i % 2)) for i in range(count)]


tokenizer = BertTokenizer.from_pretrained('../dataset/vocab.txt')


class Test_DefaultDataConverter:

    def test_with_bert_tokenizer(self):
        examples = create_test_examples('test', 10)
        converter = DefaultTaskDataConverter(tokenizer, label_list=['0', '1'])
        assert len(converter.convert(examples)) == 10


class Test_data_factory:
    def test_default_datafactory(self):
        config = TaskDataConfig()
        config.data_dir = 'build'
        datastore = LocalDataStore('test', config)
        data_processor = CSVDataProcessor('../dataset/label_text', labels=['0', '1'], train_file='train.txt',
                                          eval_file='dev.txt')
        converter = DefaultTaskDataConverter(tokenizer, label_list=['0', '1'])
        data_factory = TaskDatasetFactory(datastore, data_processor, converter)
        assert data_factory.create_train_dataset()
        assert data_factory.create_eval_dataset()





from dataclasses import dataclass
from transformers import PretrainedConfig, PreTrainedTokenizer, PreTrainedModel

from ..transformersx_base import log, join_path
from .models import task_model
from .model_config import TaskModelConfig


@dataclass
class TaskModel:
    config: PretrainedConfig
    tokenizer: PreTrainedTokenizer
    model: PreTrainedModel


class TaskModelFactory:
    def __init__(self, task_name, config: TaskModelConfig, model_class=None):
        self._task_name = task_name
        self.config = config
        self._model_class = model_class
        self._model_cache = {}

    def _create_task_model(self, model_class=None):
        t_model = task_model(model_path=self.config.model_name,
                             model_task_type=self.config.model_task_type,
                             language=self.config.language,
                             framework=self.config.framework)

        # use the custom class to replace the model Class
        if model_class is not None:
            t_model.model_class = model_class

        log.info("Built task model: {}".format(str(t_model)))
        return t_model

    def _load_task_model(self, model_base_path, model_class=None) -> TaskModel:
        t_model = self._create_task_model(model_class)
        config, tokenizer, model = t_model.load(
            num_labels=self.config.num_labels,
            unit_test=self.config.unit_test,
            cache_dir=model_base_path
        )
        log.info(
            "Loaded task model, config: {}, tokenizer: {}, "
            "model: {}".format(str(config),
                               type(tokenizer),
                               type(model))
        )

        return TaskModel(config=config, tokenizer=tokenizer, model=model)

    def _freeze_parameters(self, model):
        self._freeze_weights_main(model)

        if hasattr(model, 'num_parameters'):
            log.info("num params:" + str(model.num_parameters()))
            log.info("num trainable params:" + str(model.num_parameters(only_trainable=True)))

        if hasattr(model, 'named_parameters'):
            log.info("Model Parameters Details: ")
            for name, param in model.named_parameters():
                log.info("{}:{}".format(name, param.size()))

    def _freeze_weights_main(self, model):
        if model is None or not self.config.freeze_main: return

        main_parameters = eval("self.model." + model.main_parameter)
        if hasattr(main_parameters, "parameters"):
            for param in main_parameters.parameters():
                param.requires_grad = False

    def get_task_model(self, model_path=None, for_train=True) -> TaskModel:
        cached_name = 'pretrained' if not model_path else model_path
        task_model = self._model_cache.get(cached_name)

        if task_model: return task_model
        if not model_path:
            model_path = self.config.model_pretrained_dir

        task_model = self._load_task_model(model_path, self._model_class)

        if for_train:
            self._freeze_parameters(task_model.model)
        return task_model



from ai_harness.configuration import ComplexArguments
from ..task import TransformerTask
from ..transformersx_base import log, to_json_string


class ExampleManagement():
    TASKS = dict()

    def register_task(self, task_name, task_class):
        if self.TASKS.get(task_name) is not None:
            raise ValueError(
                "task name {} already exists, please use other name.".format(task_name)
            )
        self.TASKS.setdefault(task_name, task_class)
        if not issubclass(task_class, (TransformerTask)):
            raise ValueError(
                "task_class must be the subclass of TransformerTask"
            )
        return self

    def register_tasks(self, tasks):
        if not tasks:
            return
        for task in tasks:
            self.register_task(task[0], task[1])
        return self

    def _build_arguments(self):
        arguments = {}
        for task_name, task_class in self.TASKS.items():
            arg_obj = task_class.args_class()
            arguments[task_name] = arg_obj
        return arguments

    def start_example_task(self, args=None, test=False):
        argument_objs = self._build_arguments()

        task_name, arguments = ComplexArguments(argument_objs).parse(args)
        if test: arguments.model_args.unit_test = test
        log.info("Start Example Task:{}, with arguments:{}".format(task_name, to_json_string(arguments)))
        task_action = arguments.action
        task_instance = self.TASKS[task_name](arguments)

        eval('task_instance.' + task_action + '()')



from ..task_base import *


class SentimentTask(DefaultTransformerTask):
    def __init__(self, config: TaskConfig = None):
        super().__init__(config)

    def _create_task_context(self, config: TaskConfig) -> TaskContext:
        return TaskContext(
            task_name='sentiment',
            data_processor=CSVDataProcessor(
                data_dir=config.data_config.data_dir,
                labels=['0', '1', '2', '3']
            )
        )



from ai_transformersx.utils.text_utils import make_sentences_max_length

content = ['斯坦议会上院议长纳扎尔巴耶娃和,',
           '下面一场尼格马图林邀请全国人大常委会委员长栗战书，21号到25号对哈萨克斯坦进行正式友好访问，再努尔苏丹会见首任总统纳扎尔巴耶夫和总统托卡耶夫，与娜扎尔巴耶娃尼格马图林分别举行会谈，与总理马明共同出席相关活动。',
           '中哈',
           '关系实现',
           '中哈关系实现历史性跨越，成为国家间关系的典范。'
           ]
content2=['其实热点多方关注，在今天的新闻最关注之中，我们来关注记者张美所采写的报道，守护人与自然的和谐巍峨的大秦岭是我国南北气候的分界线，生物的多样性使其成为了地球自然资源保护西成高铁是我国首条穿越秦岭山脉的高铁，途经全国最大的连线自然保护区，当您乘坐西成高铁驰骋在秦岭山脉之间时，或许能够看见美丽的猪还飞过头顶，但是很少有人知道，这幅人与自然和谐共存的画面背后，是一支由动物保护专家组成的科研团队，日复一日年复一年的付出！']

def test_make_sentences_max_length():
    sentences = make_sentences_max_length(content2)
    print(sentences)



from .trainer import TrainerEnv, TaskTrainer, TaskTrainLoop, TaskTrainedScheduler, TaskTrainEpoch, TaskTrainStep, \
    TaskTrainerContextBase, TaskTrainerOptimizers, TaskTrainerCheckpointer, TaskTrainerEvaluator, TaskTrainerLogger, \
    TaskTrainerDataLoaders, TaskEvalPrediction, TaskPredictionOutput, TaskTrainOutput, TaskDefaultDataCollatorx, \
    TaskContext, TaskTrainerBuildContext, TaskTrainerPredictor
from .trainer_config import TaskTrainingConfig



import json

from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm
from transformers import PreTrainedModel

from .trainer_base import *
import os
from .trainer_scheduler import TaskTrainedScheduler

try:
    from torch.utils.tensorboard import SummaryWriter

    _has_tensorboard = True
except ImportError:
    try:
        from tensorboardX import SummaryWriter

        _has_tensorboard = True
    except ImportError:
        _has_tensorboard = False


def is_tensorboard_available():
    return _has_tensorboard


try:
    import wandb

    wandb.ensure_configured()
    if wandb.api.api_key is None:
        _has_wandb = False
        wandb.termwarn(
            "W&B installed but not logged in.  Run `wandb login` or set the WANDB_API_KEY env variable.")
    else:
        _has_wandb = False if os.getenv("WANDB_DISABLED") else True
except ImportError:
    _has_wandb = False


def is_wandb_available():
    return _has_wandb


@configclass
class TrainerLoggerConfig():
    logging_dir: str = field(None, "Tensorboard log dir.")
    logging_first_step: bool = field(False, "Log and eval the first global_step")
    logging_steps: int = field(500, "Log every X updates steps.")
    tpu_metrics_debug: bool = field(False, "TPU: Whether to print debug metrics")


class TaskTrainerLogger():
    def __init__(self, trainer_env: TrainerEnv,
                 model: PreTrainedModel,
                 tb_writer: Optional["SummaryWriter"] = None):
        self._env = trainer_env
        self.config: TrainerLoggerConfig = trainer_env.get_config(TrainerLoggerConfig)
        self._model = model
        self._tb_writer = tb_writer
        self.__init_logger()

    def __init_logger(self):
        if not self._tb_writer and is_tensorboard_available() and self._env.config.local_rank in [-1, 0]:
            self._tb_writer = SummaryWriter(log_dir=self.config.logging_dir)

        if not is_tensorboard_available():
            log.warning(
                "You are instantiating a Trainer but Tensorboard is not installed. You should consider installing it."
            )
        if is_wandb_available():
            self._setup_wandb()
        else:
            log.info(
                "You are instantiating a Trainer but W&B is not installed. To use wandb logging, "
                "run `pip install wandb; wandb login` see https://docs.wandb.com/huggingface."
            )

    def _setup_wandb(self):
        """
        Setup the optional Weights & Biases (`wandb`) integration.

        One can override this method to customize the setup if needed.  Find more information at https://docs.wandb.com/huggingface
        You can also override the following environment variables:

        Environment:
            WANDB_WATCH:
                (Optional, ["gradients", "all", "false"]) "gradients" by default, set to "false" to disable gradient logging
                or "all" to log gradients and parameters
            WANDB_PROJECT:
                (Optional): str - "huggingface" by default, set this to a custom string to store results in a different project
            WANDB_DISABLED:
                (Optional): boolean - defaults to false, set to "true" to disable wandb entirely
        """
        log.info('Automatic Weights & Biases logging enabled, to disable set os.environ["WANDB_DISABLED"] = "true"')
        wandb.init(project=os.getenv("WANDB_PROJECT", "huggingface"), config=vars(self._env.args))
        # keep track of model topology and gradients
        if os.getenv("WANDB_WATCH") != "false":
            wandb.watch(
                self._model, log=os.getenv("WANDB_WATCH", "gradients"), log_freq=max(100, self.config.logging_steps)
            )

    def log_pre_train(self, train_dataloader, train_scheduler: TaskTrainedScheduler):
        if self._tb_writer is not None:
            self._tb_writer.add_text("args", self._env.args.to_json_string())
            self._tb_writer.add_hparams(self._env.args.to_sanitized_dict(), metric_dict={})

        log.info("***** Running training *****")
        log.info("  Num examples = %d", self._env.num_examples(train_dataloader))
        log.info("  Num Epochs = %d", train_scheduler.num_train_epochs)
        log.info("  Instantaneous batch size per device = %d", self._env.config.per_gpu_train_batch_size)
        log.info("  Total train batch size (w. parallel, distributed & accumulation) = %d",
                 train_scheduler.total_train_batch_size)
        log.info("  Gradient Accumulation steps = %d", train_scheduler.config.gradient_accumulation_steps)
        log.info("  Total optimization steps = %d", train_scheduler.t_total)

    def is_need_log_step(self, global_step):
        return (self.config.logging_steps > 0 and global_step % self.config.logging_steps == 0) or (
                global_step == 1 and self.config.logging_first_step)

    def log_train_step(self, epoch,
                       global_step,
                       logs: Dict[str, float],
                       iterator: Optional[tqdm] = None) -> None:
        if epoch is not None:  logs["epoch"] = epoch
        if self._tb_writer:
            for k, v in logs.items():
                self._tb_writer.add_scalar(k, v, global_step)
        if is_wandb_available(): wandb.log(logs, step=global_step)
        output = json.dumps({**logs, **{"step": global_step}})

        (iterator.write(output) if iterator else print(output))

    def log_train_epoch(self):
        if self.config.tpu_metrics_debug:
            # tpu-comment: Logging debug metrics for PyTorch/XLA (compile, execute times, ops, etc.)
            self._env.tpu_metrics()

    def log_train_end(self):
        if self._tb_writer: self._tb_writer.close()

        log.info("\n\nTraining completed. Do not forget to share your model on huggingface.co/models =)\n\n")
